from datetime import datetime, timedelta
from pathlib import Path

import pytest

from cg.constants import Pipeline
from cg.constants.constants import FileFormat
from cg.io.controller import ReadFile
from cg.meta.report.balsamic import BalsamicReportAPI
from cg.meta.report.mip_dna import MipDNAReportAPI
from cg.meta.report.rnafusion import RnafusionReportAPI
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import Case
from tests.apps.scout.conftest import MockScoutApi
from tests.mocks.balsamic_analysis_mock import MockBalsamicAnalysis
from tests.mocks.limsmock import MockLimsAPI
from tests.mocks.mip_analysis_mock import MockMipAnalysis
from tests.mocks.report import MockChanjo, MockHousekeeperMipDNAReportAPI


@pytest.fixture(scope="function")
def report_api_mip_dna(
    cg_context: CGConfig, lims_samples: list[dict], report_store: Store
) -> MipDNAReportAPI:
    """MIP DNA ReportAPI fixture."""
    cg_context.meta_apis["analysis_api"] = MockMipAnalysis(
        config=cg_context, pipeline=Pipeline.MIP_DNA
    )
    cg_context.status_db_ = report_store
    cg_context.lims_api_ = MockLimsAPI(cg_context, lims_samples)
    cg_context.chanjo_api_ = MockChanjo()
    cg_context.scout_api_ = MockScoutApi(cg_context)
    return MockHousekeeperMipDNAReportAPI(cg_context, cg_context.meta_apis["analysis_api"])


@pytest.fixture(scope="function")
def report_api_balsamic(
    cg_context: CGConfig, lims_samples: list[dict], report_store: Store
) -> BalsamicReportAPI:
    """BALSAMIC ReportAPI fixture."""
    cg_context.meta_apis["analysis_api"] = MockBalsamicAnalysis(cg_context)
    cg_context.status_db_ = report_store
    cg_context.lims_api_ = MockLimsAPI(cg_context, lims_samples)
    cg_context.scout_api_ = MockScoutApi(cg_context)
    return BalsamicReportAPI(cg_context, cg_context.meta_apis["analysis_api"])


@pytest.fixture(scope="function")
def report_api_rnafusion(
    rnafusion_context: CGConfig, lims_samples: list[dict]
) -> RnafusionReportAPI:
    """Rnafusion report API fixture."""
    rnafusion_context.lims_api_ = MockLimsAPI(rnafusion_context, lims_samples)
    rnafusion_context.scout_api_ = MockScoutApi(rnafusion_context)
    return RnafusionReportAPI(rnafusion_context, rnafusion_context.meta_apis["analysis_api"])


@pytest.fixture(scope="function")
def case_mip_dna(case_id: str, report_api_mip_dna: MipDNAReportAPI) -> Case:
    """MIP DNA case instance."""
    return report_api_mip_dna.status_db.get_case_by_internal_id(internal_id=case_id)


@pytest.fixture(scope="function")
def case_balsamic(case_id: str, report_api_balsamic: BalsamicReportAPI) -> Case:
    """BALSAMIC case instance."""
    return report_api_balsamic.status_db.get_case_by_internal_id(internal_id=case_id)


@pytest.fixture(scope="function")
def case_samples_data(case_id: str, report_api_mip_dna: MipDNAReportAPI):
    """MIP DNA family sample object."""
    return report_api_mip_dna.status_db.get_case_samples_by_case_id(case_internal_id=case_id)


@pytest.fixture(scope="function")
def mip_analysis_api(cg_context: CGConfig) -> MockMipAnalysis:
    """MIP analysis mock data."""
    return MockMipAnalysis(config=cg_context, pipeline=Pipeline.MIP_DNA)


@pytest.fixture(scope="session")
def lims_family(fixtures_dir: Path) -> dict:
    """Returns a lims-like case of samples."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=Path(fixtures_dir, "report", "lims_family.json")
    )


@pytest.fixture(scope="session")
def lims_samples(lims_family: dict) -> list[dict]:
    """Returns the samples of a lims case."""
    return lims_family["samples"]


@pytest.fixture(scope="function", autouse=True)
def report_store(analysis_store, helpers, timestamp_yesterday):
    """A mock store instance for report testing."""
    case = analysis_store.get_cases()[0]
    helpers.add_analysis(
        analysis_store, case, pipeline=Pipeline.MIP_DNA, started_at=timestamp_yesterday
    )
    helpers.add_analysis(analysis_store, case, pipeline=Pipeline.MIP_DNA, started_at=datetime.now())
    # Mock sample dates to calculate processing times
    for family_sample in analysis_store.get_case_samples_by_case_id(
        case_internal_id=case.internal_id
    ):
        family_sample.sample.ordered_at = timestamp_yesterday - timedelta(days=2)
        family_sample.sample.received_at = timestamp_yesterday - timedelta(days=1)
        family_sample.sample.prepared_at = timestamp_yesterday
        family_sample.sample.last_sequenced_at = timestamp_yesterday
        family_sample.sample.delivered_at = datetime.now()
    return analysis_store


@pytest.fixture(scope="session")
def rnafusion_validated_metrics() -> dict[str, str]:
    """Return Rnafusion raw analysis metrics dictionary."""
    return {
        "gc_content": "51.7",
        "ribosomal_bases": "65.81",
        "q20_rate": "97.48",
        "q30_rate": "92.95",
        "mapped_reads": "96.53",
        "rin": "10.0",
        "input_amount": "300.0",
        "insert_size": "N/A",
        "insert_size_peak": "N/A",
        "mean_length_r1": "99.0",
        "million_read_pairs": "75.0",
        "bias_5_3": "1.12",
        "pct_adapter": "12.01",
        "duplicates": "14.86",
        "mrna_bases": "85.97",
        "pct_surviving": "99.42",
        "uniquely_mapped_reads": "91.02",
    }



import datetime
from unittest import mock

import pytest
from click.testing import CliRunner
from housekeeper.store.models import Archive, File
from requests import HTTPError, Response

from cg.cli.archive import archive_spring_files, delete_file, update_job_statuses
from cg.constants import EXIT_SUCCESS, SequencingFileTag
from cg.constants.archiving import ArchiveLocations
from cg.io.controller import APIRequest
from cg.meta.archive.ddn.constants import (
    FAILED_JOB_STATUSES,
    ONGOING_JOB_STATUSES,
    JobStatus,
)
from cg.meta.archive.ddn.ddn_data_flow_client import DDNDataFlowClient
from cg.meta.archive.ddn.models import (
    AuthToken,
    GetJobStatusPayload,
    GetJobStatusResponse,
    TransferPayload,
    TransferResponse,
)
from cg.models.cg_config import CGConfig


def test_limit_and_archive_all_fails(cli_runner: CliRunner, cg_context: CGConfig):
    """Tests that when invoking archive-spring-files in the Archive CLI module, the command is
    aborted if both a limit and the --archive-all flag is provided."""

    # GIVEN a CLI runner and a context

    # WHEN invoking archive_spring_files with both a given limit and specifying archive_all
    result = cli_runner.invoke(
        archive_spring_files,
        ["--limit", 100, "--archive-all"],
        obj=cg_context,
    )

    # THEN the command should have exited with an exit_code 1
    assert result.exit_code == 1
    assert (
        "Incorrect input parameters - please do not provide both a limit and set --archive-all."
        in result.stdout
    )


def test_archive_spring_files_success(
    cli_runner: CliRunner,
    archive_context: CGConfig,
    archival_job_id: int,
    test_auth_token: AuthToken,
    ok_miria_response: Response,
):
    """Tests that the CLI command 'cg archive archive-spring-files' adds an archive with the correct
    job_id when an archiving job is launched via Miria."""

    # GIVEN a CLI runner and a context

    # GIVEN a spring file belonging to a customer with archive location 'karolinska_bucket'
    all_non_archived_spring_files: list[
        File
    ] = archive_context.housekeeper_api.get_non_archived_spring_files()
    assert len(all_non_archived_spring_files) == 1
    spring_file: File = all_non_archived_spring_files[0]

    # WHEN running 'cg archive archive-spring-files'
    with mock.patch.object(
        AuthToken,
        "model_validate",
        return_value=test_auth_token,
    ), mock.patch.object(
        APIRequest,
        "api_request_from_content",
        return_value=ok_miria_response,
    ), mock.patch.object(
        TransferPayload, "post_request", return_value=TransferResponse(jobId=archival_job_id)
    ):
        result = cli_runner.invoke(
            archive_spring_files,
            obj=archive_context,
        )

    # THEN the command should have executed without fail and added an entry in the archive with the archiving task id
    # returned by Miria
    assert result.exit_code == EXIT_SUCCESS
    assert spring_file.archive.archiving_task_id == archival_job_id


@pytest.mark.parametrize(
    "job_status", [JobStatus.COMPLETED, FAILED_JOB_STATUSES[0], ONGOING_JOB_STATUSES[0]]
)
def test_get_archival_job_status(
    cli_runner: CliRunner,
    archive_context: CGConfig,
    archival_job_id: int,
    test_auth_token: AuthToken,
    ok_miria_response: Response,
    job_status: JobStatus,
):
    """Tests that when invoking update_job_statuses in the Archive CLI module with an ongoing archival job,
    the database is updated according to whether the job has completed, failed or is still ongoing.
    """

    # GIVEN a CLI runner and a context

    # GIVEN an archive entry with an ongoing archival
    assert len(archive_context.housekeeper_api.get_archive_entries()) == 1
    assert not archive_context.housekeeper_api.get_archive_entries()[0].archived_at

    # WHEN invoking update_job_statuses
    with mock.patch.object(
        AuthToken,
        "model_validate",
        return_value=test_auth_token,
    ), mock.patch.object(
        APIRequest,
        "api_request_from_content",
        return_value=ok_miria_response,
    ), mock.patch.object(
        GetJobStatusPayload,
        "get_job_status",
        return_value=GetJobStatusResponse(id=archival_job_id, status=job_status),
    ):
        result = cli_runner.invoke(
            update_job_statuses,
            obj=archive_context,
        )

    # THEN the command should have exited successfully and updated the archive record
    assert result.exit_code == 0
    if job_status == JobStatus.COMPLETED:
        assert archive_context.housekeeper_api.get_archive_entries(
            archival_task_id=archival_job_id
        )[0].archived_at
    elif job_status in FAILED_JOB_STATUSES:
        assert not archive_context.housekeeper_api.get_archive_entries(
            archival_task_id=archival_job_id
        )
    elif job_status in ONGOING_JOB_STATUSES:
        assert not archive_context.housekeeper_api.get_archive_entries(
            archival_task_id=archival_job_id
        )[0].archived_at


@pytest.mark.parametrize(
    "job_status", [JobStatus.COMPLETED, FAILED_JOB_STATUSES[0], ONGOING_JOB_STATUSES[0]]
)
def test_get_retrieval_job_status(
    cli_runner: CliRunner,
    archive_context: CGConfig,
    retrieval_job_id: int,
    test_auth_token: AuthToken,
    ok_miria_response: Response,
    job_status: JobStatus,
):
    """Tests that when invoking update_job_statuses in the Archive CLI module with an ongoing retrieval job,
    the database is updated according to whether the job has completed, failed or is still ongoing.
    """

    # GIVEN a CLI runner and a context

    # GIVEN an archive entry with an ongoing retrieval
    retrieving_archive: Archive = archive_context.housekeeper_api.get_archive_entries()[0]
    retrieving_archive.archived_at = datetime.datetime.now()
    retrieving_archive.retrieval_task_id = retrieval_job_id

    # WHEN invoking update_job_statuses
    with mock.patch.object(
        AuthToken,
        "model_validate",
        return_value=test_auth_token,
    ), mock.patch.object(
        APIRequest,
        "api_request_from_content",
        return_value=ok_miria_response,
    ), mock.patch.object(
        GetJobStatusPayload,
        "get_job_status",
        return_value=GetJobStatusResponse(id=retrieval_job_id, status=job_status),
    ):
        result = cli_runner.invoke(
            update_job_statuses,
            obj=archive_context,
        )

    # THEN the command should have exited successfully and updated the archive record
    assert result.exit_code == 0
    if job_status == JobStatus.COMPLETED:
        assert archive_context.housekeeper_api.get_archive_entries(
            retrieval_task_id=retrieval_job_id
        )[0].retrieved_at
    elif job_status in FAILED_JOB_STATUSES:
        assert not archive_context.housekeeper_api.get_archive_entries(
            retrieval_task_id=retrieval_job_id
        )
    elif job_status in ONGOING_JOB_STATUSES:
        assert not archive_context.housekeeper_api.get_archive_entries(
            retrieval_task_id=retrieval_job_id
        )[0].retrieved_at


def test_delete_file_raises_http_error(
    cli_runner: CliRunner,
    archive_context: CGConfig,
    failed_delete_file_response: Response,
    test_auth_token: AuthToken,
    archival_job_id: int,
):
    """Tests that an HTTP error is raised when the Miria response is unsuccessful for a delete file request,
    and that the file is not removed from Housekeeper."""

    # GIVEN a spring file which is archived via Miria
    spring_file: File = archive_context.housekeeper_api.files(
        tags={SequencingFileTag.SPRING, ArchiveLocations.KAROLINSKA_BUCKET}
    ).first()
    spring_file_path: str = spring_file.path
    if not spring_file.archive:
        archive_context.housekeeper_api.add_archives(
            files=[spring_file], archive_task_id=archival_job_id
        )
    archive_context.housekeeper_api.set_archive_archived_at(
        file_id=spring_file.id, archiving_task_id=archival_job_id
    )

    # GIVEN that the request returns a failed response
    with mock.patch.object(
        DDNDataFlowClient,
        "_get_auth_token",
        return_value=test_auth_token,
    ), mock.patch.object(
        APIRequest,
        "api_request_from_content",
        return_value=failed_delete_file_response,
    ), pytest.raises(
        HTTPError
    ):
        # WHEN trying to delete the file via Miria and in Housekeeper

        # THEN an HTTPError should be raised
        cli_runner.invoke(
            delete_file, [spring_file.path], obj=archive_context, catch_exceptions=False
        )

    # THEN the file should still be in Housekeeper
    assert archive_context.housekeeper_api.files(path=spring_file_path)


def test_delete_file_success(
    cli_runner: CliRunner,
    archive_context: CGConfig,
    ok_delete_file_response: Response,
    test_auth_token: AuthToken,
    archival_job_id: int,
):
    """Tests that given a successful response from Miria, the file is deleted and removed from Housekeeper."""

    # GIVEN a spring file which is archived via Miria
    spring_file: File = archive_context.housekeeper_api.files(
        tags={SequencingFileTag.SPRING, ArchiveLocations.KAROLINSKA_BUCKET}
    ).first()
    spring_file_id: int = spring_file.id
    if not spring_file.archive:
        archive_context.housekeeper_api.add_archives(
            files=[spring_file], archive_task_id=archival_job_id
        )
    archive_context.housekeeper_api.set_archive_archived_at(
        file_id=spring_file.id, archiving_task_id=archival_job_id
    )

    # GIVEN that the delete request returns a successful response
    with mock.patch.object(
        DDNDataFlowClient,
        "_get_auth_token",
        return_value=test_auth_token,
    ), mock.patch.object(
        APIRequest,
        "api_request_from_content",
        return_value=ok_delete_file_response,
    ):
        # WHEN trying to delete the file via Miria and in Housekeeper

        # THEN no error is raised
        cli_runner.invoke(
            delete_file, [spring_file.path], obj=archive_context, catch_exceptions=False
        )

    # THEN the file is removed from Housekeeper
    assert not archive_context.housekeeper_api.get_file(spring_file_id)



from pathlib import Path

from cg.apps.sequencing_metrics_parser.parsers.bcl2fastq_to_sequencing_statistics import (
    create_bcl2fastq_metrics,
    create_bcl2fastq_undetermined_metrics,
)
from cg.apps.sequencing_metrics_parser.parsers.bcl_convert_to_sequencing_statistics import (
    create_bcl_convert_undetermined_metrics,
    create_sample_lane_sequencing_metrics_from_bcl_convert_metrics_for_flow_cell,
)
from cg.constants.demultiplexing import BclConverter
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData
from cg.store.models import SampleLaneSequencingMetrics


def create_sample_lane_sequencing_metrics_for_flow_cell(
    flow_cell_directory: Path, bcl_converter: str
) -> list[SampleLaneSequencingMetrics]:
    """Parse the sequencing metrics data for the correct demultiplexing software
    into the sequencing statistics model."""
    if bcl_converter == BclConverter.BCLCONVERT or bcl_converter == BclConverter.DRAGEN:
        return create_sample_lane_sequencing_metrics_from_bcl_convert_metrics_for_flow_cell(
            flow_cell_directory
        )
    return create_bcl2fastq_metrics(flow_cell_directory)


def create_undetermined_non_pooled_metrics(
    flow_cell: FlowCellDirectoryData,
) -> list[SampleLaneSequencingMetrics]:
    """Return sequencing metrics for any undetermined reads in non-pooled lanes."""

    non_pooled_lanes_and_samples: list[
        tuple[int, str]
    ] = flow_cell.sample_sheet.get_non_pooled_lanes_and_samples()

    if flow_cell.bcl_converter == BclConverter.BCL2FASTQ:
        return create_bcl2fastq_undetermined_metrics(
            bcl2fastq_flow_cell_path=flow_cell.path,
            non_pooled_lane_sample_pairs=non_pooled_lanes_and_samples,
        )
    return create_bcl_convert_undetermined_metrics(
        flow_cell_dir=flow_cell.path,
        non_pooled_lane_sample_pairs=non_pooled_lanes_and_samples,
    )



"""Constants specific for compression."""
import datetime

# Constants for crunchy
FASTQ_FIRST_READ_SUFFIX: str = "_R1_001.fastq.gz"
FASTQ_SECOND_READ_SUFFIX: str = "_R2_001.fastq.gz"
FLAG_PATH_SUFFIX: str = ".crunchy.txt"
MAX_READS_PER_GB: int = 18_000_000
CRUNCHY_MIN_GB_PER_PROCESS: int = 30
PENDING_PATH_SUFFIX: str = ".crunchy.pending.txt"


# Number of days until FASTQs counts as old
FASTQ_DELTA = 21
FASTQ_DATETIME_DELTA = datetime.timedelta(days=FASTQ_DELTA)



"""Tests for functions related to indexes."""
import pytest

from cg.apps.demultiplex.sample_sheet.index import (
    Index,
    get_hamming_distance_for_indexes,
    get_reverse_complement_dna_seq,
    get_valid_indexes,
    is_padding_needed,
)


def test_get_valid_indexes():
    """Test that the function get_valid_indexes returns a list of Index objects."""
    # GIVEN a sample sheet API

    # WHEN fetching the indexes
    indexes: list[Index] = get_valid_indexes()

    # THEN assert that the indexes are correct
    assert indexes
    assert isinstance(indexes[0], Index)


@pytest.mark.parametrize(
    "index1_cycles, index2_cycles, sample_index_length, expected_is_padding_needed",
    [
        (10, 10, 8, True),
        (10, 10, 10, False),
        (17, 8, 17, False),
        (8, 8, 10, False),
        (8, 8, 8, False),
    ],
)
def test_is_padding_needed(
    index1_cycles: int,
    index2_cycles: int,
    sample_index_length: int,
    expected_is_padding_needed: bool,
):
    """Test that evaluating if a situation needs padding returns the expected value."""
    # GIVEN a sample index length and the number of index cycles reads stated in the run parameters

    # WHEN checking if padding is needed
    padding_needed: bool = is_padding_needed(
        index1_cycles=index1_cycles,
        index2_cycles=index2_cycles,
        sample_index_length=sample_index_length,
    )

    # THEN assert that the result is the expected
    assert padding_needed == expected_is_padding_needed


def test_get_reverse_complement():
    """Test that getting the reverse complement of a DNA strain returns the correct sequence."""
    # GIVEN a DNA strain and its reverse complement
    dna_strain: str = "ACCTCTGT"
    reversed_complement: str = "ACAGAGGT"

    # WHEN getting the reverse complement of the DNA strain
    returned_reverse_complement: str = get_reverse_complement_dna_seq(dna=dna_strain)

    # THEN the result is the expected
    assert returned_reverse_complement == reversed_complement


def test_get_reverse_complement_not_dna(caplog):
    """Test that using a string without 'A', 'C', 'G' or 'T' fails."""
    # GIVEN a non-DNA strain
    strain: str = "ACCUCTGU"

    # WHEN getting the reverse complement
    with pytest.raises(KeyError):
        # THEN the process fails due to not recognising the unknown nucleotide
        get_reverse_complement_dna_seq(dna=strain)


def test_get_hamming_distance_index_1_different_lengths():
    """Test that hamming distance between indexes with same prefix but different lengths is zero."""
    # GIVEN two index_1 sequences with the same prefixes but different lengths
    sequence_1: str = "GATTACA"
    sequence_2: str = "GATTACAXX"

    # WHEN getting the hamming distance between them in any order

    # THEN the distance is zero
    assert get_hamming_distance_for_indexes(sequence_1=sequence_1, sequence_2=sequence_2) == 0
    assert get_hamming_distance_for_indexes(sequence_1=sequence_2, sequence_2=sequence_1) == 0

    # WHEN getting the hamming distance between themselves

    # THEN the distance is zero
    assert get_hamming_distance_for_indexes(sequence_1=sequence_1, sequence_2=sequence_1) == 0
    assert get_hamming_distance_for_indexes(sequence_1=sequence_2, sequence_2=sequence_2) == 0


def test_get_hamming_distance_index_1_different_prefixes():
    """Test that hamming distance for index 1 counts different characters from the left."""
    # GIVEN two index_1 sequences with different lengths differing by two characters
    # when aligned to the left
    sequence_1: str = "GATXX"
    sequence_2: str = "GATTACA"

    # WHEN getting the hamming distance between them in any order

    # THEN the distance is equal to the number of different characters
    assert get_hamming_distance_for_indexes(sequence_1=sequence_1, sequence_2=sequence_2) == 2
    assert get_hamming_distance_for_indexes(sequence_1=sequence_2, sequence_2=sequence_1) == 2



from pathlib import Path

import pytest

from cg.constants.constants import FileFormat
from cg.io.controller import ReadFile


@pytest.fixture(name="balsamic_config_path")
def balsamic_config_path(fixtures_dir) -> Path:
    """Returns path to BALSAMIC case_config.json"""

    return Path(fixtures_dir, "apps", "balsamic", "case", "config.json")


@pytest.fixture(name="balsamic_metrics_path")
def balsamic_metrics_path(fixtures_dir) -> Path:
    """Returns path to BALSAMIC case_metrics_deliverables.yaml"""

    return Path(fixtures_dir, "apps", "balsamic", "case", "metrics_deliverables.yaml")


@pytest.fixture(name="balsamic_config_raw")
def balsamic_config(balsamic_config_path) -> dict:
    """Return BALSAMIC config file as a dictionary"""

    config: dict = ReadFile.get_content_from_file(
        file_format=FileFormat.YAML, file_path=balsamic_config_path
    )
    return config


@pytest.fixture(name="balsamic_metrics_raw")
def balsamic_metrics(balsamic_metrics_path) -> dict:
    """Return BALSAMIC metrics file as a dictionary"""

    metrics: dict = ReadFile.get_content_from_file(
        file_format=FileFormat.YAML, file_path=balsamic_metrics_path
    )
    return metrics



"""Nextflow related constants."""
from enum import IntEnum

NFX_WORK_DIR = "work"
NFX_SAMPLE_HEADER = "sample"
NFX_READ1_HEADER = "fastq_1"
NFX_READ2_HEADER = "fastq_2"
NFX_SAMPLESHEET_READS_HEADERS = [NFX_READ1_HEADER, NFX_READ2_HEADER]
NFX_SAMPLESHEET_HEADERS = [NFX_SAMPLE_HEADER] + NFX_SAMPLESHEET_READS_HEADERS
DELIVER_FILE_HEADERS = ["format", "id", "path", "path_index", "step", "tag"]
NXF_PID_FILE_ENV = "NXF_PID_FILE"
NXF_JVM_ARGS_ENV = "NXF_JVM_ARGS"
JAVA_MEMORY_HEADJOB = "-Xmx5g"


class SlurmHeadJobDefaults(IntEnum):
    """Default parameters for slurm head jobs."""

    HOURS: int = 96
    MEMORY: int = 10
    NUMBER_TASKS: int = 1



from pathlib import Path

from cg.models.cg_config import EmailBaseSettings


class EmailInfo(EmailBaseSettings):
    receiver_email: str
    sender_email: str = "hiseq.clinical@hasta.scilifelab.se"
    subject: str
    message: str
    file: Path | None



import logging

import click

from cg.constants import STATUS_OPTIONS, DataDelivery, Pipeline, Priority
from cg.constants.archiving import PDC_ARCHIVE_LOCATION
from cg.constants.subject import Sex
from cg.meta.transfer.external_data import ExternalDataAPI
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import (
    Application,
    ApplicationVersion,
    Case,
    CaseSample,
    Collaboration,
    Customer,
    Panel,
    Sample,
    User,
)
from cg.utils.click.EnumChoice import EnumChoice

LOG = logging.getLogger(__name__)


@click.group()
def add():
    """Add new things to the database."""
    pass


@add.command("customer")
@click.argument("internal_id")
@click.argument("name")
@click.option(
    "-cg",
    "--collaboration-internal-ids",
    "collaboration_internal_ids",
    multiple=True,
    help="List of internal IDs for the collaborations the customer should belong to",
)
@click.option(
    "-ia",
    "--invoice-address",
    "invoice_address",
    required=True,
    help="Street address, Post code, City",
)
@click.option(
    "-ir",
    "--invoice-reference",
    "invoice_reference",
    required=True,
    help="Invoice reference (text)",
)
@click.option(
    "-da",
    "--data-archive-location",
    "data_archive_location",
    help="Specifies where to store data for the customer.",
    default=PDC_ARCHIVE_LOCATION,
    show_default=True,
    required=False,
)
@click.option(
    "-ic",
    "--is-clinical",
    "is_clinical",
    help="Set to true to indicate that this customer is clinical"
    " and handle data storage accordingly, i.e. for 10 years instead if 2 years.",
    is_flag=True,
    default=False,
    show_default=True,
    required=False,
)
@click.pass_obj
def add_customer(
    context: CGConfig,
    internal_id: str,
    name: str,
    collaboration_internal_ids: list[str] | None,
    invoice_address: str,
    invoice_reference: str,
    data_archive_location: str,
    is_clinical: bool,
):
    """Add a new customer with a unique internal id and name."""
    collaboration_internal_ids = collaboration_internal_ids or []
    status_db: Store = context.status_db

    existing_customer: Customer = status_db.get_customer_by_internal_id(
        customer_internal_id=internal_id
    )
    if existing_customer:
        LOG.error(f"{existing_customer.name}: customer already added")
        raise click.Abort

    collaborations: list[Collaboration] = [
        status_db.get_collaboration_by_internal_id(internal_id=collaboration_internal_id)
        for collaboration_internal_id in collaboration_internal_ids
    ]

    new_customer: Customer = status_db.add_customer(
        internal_id=internal_id,
        name=name,
        invoice_address=invoice_address,
        invoice_reference=invoice_reference,
        data_archive_location=data_archive_location,
        is_clinical=is_clinical,
    )
    for collaboration in collaborations:
        new_customer.collaborations.append(collaboration)
    status_db.session.add(new_customer)
    status_db.session.commit()
    message: str = f"customer added: {new_customer.internal_id} ({new_customer.id})"
    LOG.info(message)


@add.command("user")
@click.option("-a", "--admin", is_flag=True, help="make the user an admin")
@click.option(
    "-c",
    "--customer",
    "customer_id",
    required=True,
    help="internal ID for the customer of the user",
)
@click.argument("email")
@click.argument("name")
@click.pass_obj
def add_user(context: CGConfig, admin: bool, customer_id: str, email: str, name: str):
    """Add a new user with an EMAIL (login) and a NAME (full)."""
    status_db: Store = context.status_db

    customer_obj: Customer = status_db.get_customer_by_internal_id(customer_internal_id=customer_id)
    existing_user: User = status_db.get_user_by_email(email=email)
    if existing_user:
        LOG.error(f"{existing_user.name}: user already added")
        raise click.Abort

    new_user: User = status_db.add_user(
        customer=customer_obj, email=email, name=name, is_admin=admin
    )
    status_db.session.add(new_user)
    status_db.session.commit()
    LOG.info(f"User added: {new_user.email} ({new_user.id})")


@add.command("sample")
@click.option("-l", "--lims", "lims_id", help="LIMS id for the sample")
@click.option(
    "-d", "--down-sampled", type=int, help="How many reads is the sample down sampled to?"
)
@click.option("-o", "--order", help="Name of the order the sample belongs to")
@click.option(
    "-s",
    "--sex",
    type=EnumChoice(Sex, use_value=False),
    required=True,
    help="Sample pedigree sex",
)
@click.option("-a", "--application-tag", required=True, help="application tag name")
@click.option(
    "-p",
    "--priority",
    type=EnumChoice(Priority, use_value=False),
    default="standard",
    help="Set the priority for the samples",
)
@click.argument("customer_id")
@click.argument("name")
@click.pass_obj
def add_sample(
    context: CGConfig,
    lims_id: str | None,
    down_sampled: int | None,
    sex: Sex,
    order: str | None,
    application_tag: str,
    priority: Priority,
    customer_id: str,
    name: str,
):
    """Add a sample for CUSTOMER_ID with a NAME (display)."""
    status_db: Store = context.status_db

    customer: Customer = status_db.get_customer_by_internal_id(customer_internal_id=customer_id)
    if not customer:
        LOG.error(f"Customer: {customer_id} not found")
        raise click.Abort
    application: Application = status_db.get_application_by_tag(tag=application_tag)
    if not application:
        LOG.error(f"Application: {application_tag} not found")

        raise click.Abort
    new_record: Sample = status_db.add_sample(
        name=name,
        sex=sex,
        downsampled_to=down_sampled,
        internal_id=lims_id,
        order=order,
        priority=priority,
    )
    new_record.application_version: ApplicationVersion = (
        status_db.get_current_application_version_by_tag(tag=application_tag)
    )
    new_record.customer: Customer = customer
    status_db.session.add(new_record)
    status_db.session.commit()
    LOG.info(f"{new_record.internal_id}: new sample added")


@add.command("case")
@click.option(
    "--priority",
    type=EnumChoice(Priority, use_value=False),
    default="standard",
    help="analysis priority",
)
@click.option("-p", "--panel", "panel_abbreviations", multiple=True, help="default gene panels")
@click.option(
    "-a",
    "--analysis",
    "data_analysis",
    help="Analysis workflow",
    required=True,
    type=EnumChoice(Pipeline),
)
@click.option(
    "-dd",
    "--data-delivery",
    "data_delivery",
    help="Update case data delivery",
    required=True,
    type=EnumChoice(DataDelivery),
)
@click.option("-t", "--ticket", help="Ticket number", required=True)
@click.argument("customer_id")
@click.argument("name")
@click.pass_obj
def add_case(
    context: CGConfig,
    priority: Priority,
    panel_abbreviations: tuple[str],
    data_analysis: Pipeline,
    data_delivery: DataDelivery,
    customer_id: str,
    name: str,
    ticket: str,
):
    """Add a case with the given name and associated with the given customer"""
    status_db: Store = context.status_db

    customer: Customer = status_db.get_customer_by_internal_id(customer_internal_id=customer_id)
    if customer is None:
        LOG.error(f"{customer_id}: customer not found")
        raise click.Abort

    for panel_abbreviation in panel_abbreviations:
        panel: Panel = status_db.get_panel_by_abbreviation(abbreviation=panel_abbreviation)

        if panel is None:
            LOG.error(f"{panel_abbreviation}: panel not found")
            raise click.Abort

    new_case: Case = status_db.add_case(
        data_analysis=data_analysis,
        data_delivery=data_delivery,
        name=name,
        panels=list(panel_abbreviations),
        priority=priority,
        ticket=ticket,
    )

    new_case.customer: Customer = customer
    status_db.session.add(new_case)
    status_db.session.commit()
    LOG.info(f"{new_case.internal_id}: new case added")


@add.command("relationship")
@click.option("-m", "--mother-id", help="Sample ID for mother of sample")
@click.option("-f", "--father-id", help="Sample ID for father of sample")
@click.option("-s", "--status", type=click.Choice(STATUS_OPTIONS), required=True)
@click.argument("case-id")
@click.argument("sample-id")
@click.pass_obj
def link_sample_to_case(
    context: CGConfig,
    mother_id: str | None,
    father_id: str | None,
    status: str,
    case_id: str,
    sample_id: str,
):
    """Create a link between a case id and a sample id."""
    status_db: Store = context.status_db
    mother: Sample | None = None
    father: Sample | None = None
    case_obj: Case = status_db.get_case_by_internal_id(internal_id=case_id)
    if case_obj is None:
        LOG.error(f"{case_id}: family not found")
        raise click.Abort

    sample: Sample = status_db.get_sample_by_internal_id(internal_id=sample_id)
    if sample is None:
        LOG.error(f"{sample_id}: sample not found")
        raise click.Abort

    if mother_id:
        mother: Sample = status_db.get_sample_by_internal_id(internal_id=mother_id)
        if mother is None:
            LOG.error(f"{mother_id}: mother not found")
            raise click.Abort

    if father_id:
        father: Sample = status_db.get_sample_by_internal_id(internal_id=father_id)
        if father is None:
            LOG.error(f"{father_id}: father not found")
            raise click.Abort

    new_record: CaseSample = status_db.relate_sample(
        case=case_obj, sample=sample, status=status, mother=mother, father=father
    )
    status_db.session.add(new_record)
    status_db.session.commit()
    LOG.info(f"related {case_id} to {sample_id}")


@add.command("external")
@click.option(
    "-t",
    "--ticket",
    type=str,
    help="Ticket id",
    required=True,
)
@click.option("--dry-run", is_flag=True)
@click.pass_obj
def download_external_delivery_data_to_hpc(context: CGConfig, ticket: str, dry_run: bool):
    """Downloads external data from the delivery server and places it in appropriate folder on
    the HPC"""
    external_data_api = ExternalDataAPI(config=context)
    external_data_api.transfer_sample_files_from_source(ticket=ticket, dry_run=dry_run)


@add.command("external-hk")
@click.option(
    "-t",
    "--ticket",
    type=str,
    help="Ticket id",
    required=True,
)
@click.option("--dry-run", is_flag=True)
@click.option(
    "--force", help="Overwrites any any previous samples in the customer directory", is_flag=True
)
@click.pass_obj
def add_external_data_to_hk(context: CGConfig, ticket: str, dry_run: bool, force):
    """Adds external data to Housekeeper"""
    external_data_api = ExternalDataAPI(config=context)
    external_data_api.add_transfer_to_housekeeper(dry_run=dry_run, ticket=ticket, force=force)



from pathlib import Path

import pytest

from cg.constants import FileExtensions


@pytest.fixture(scope="session")
def deliverables_template_content() -> list[dict]:
    return [
        {
            "format": "yml",
            "id": "CASEID",
            "path": Path("PATHTOCASE", "pipeline_info", "software_versions.yml").as_posix(),
            "path_index": None,
            "step": "software-versions",
            "tag": "software-versions",
        },
        {
            "format": "json",
            "id": "CASEID",
            "path": Path("PATHTOCASE", "multiqc", "multiqc_data", "multiqc_data")
            .with_suffix(FileExtensions.JSON)
            .as_posix(),
            "path_index": None,
            "step": "multiqc-json",
            "tag": "multiqc-json",
        },
    ]



"""Code for store part of CLI."""

import logging

import click

from cg.apps.crunchy.crunchy import CrunchyAPI
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.meta.compress.compress import CompressAPI
from cg.models.cg_config import CGConfig

from .fastq import (
    store_case,
    store_demultiplexed_flow_cell,
    store_flow_cell,
    store_sample,
    store_ticket,
)

LOG = logging.getLogger(__name__)


@click.group()
@click.pass_obj
def store(context: CGConfig):
    """Command for storing files."""
    LOG.info("Running CG store command")
    housekeeper_api: HousekeeperAPI = context.housekeeper_api
    crunchy_api: CrunchyAPI = context.crunchy_api

    compress_api = CompressAPI(
        hk_api=housekeeper_api, crunchy_api=crunchy_api, demux_root=context.flow_cells_dir
    )
    context.meta_apis["compress_api"] = compress_api


for sub_cmd in [
    store_case,
    store_demultiplexed_flow_cell,
    store_flow_cell,
    store_sample,
    store_ticket,
]:
    store.add_command(sub_cmd)



import logging
from pathlib import Path

from cg.apps.gt import GenotypeAPI
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.housekeeper.hk import models as housekeeper_models
from cg.constants.constants import FileFormat, Pipeline, PrepCategory
from cg.constants.housekeeper_tags import HkMipAnalysisTag
from cg.constants.subject import Sex
from cg.io.controller import ReadFile
from cg.models.mip.mip_metrics_deliverables import MIPMetricsDeliverables
from cg.store.models import Analysis, Case, Sample

LOG = logging.getLogger(__name__)


class UploadGenotypesAPI(object):
    def __init__(
        self,
        hk_api: HousekeeperAPI,
        gt_api: GenotypeAPI,
    ):
        LOG.info("Initializing UploadGenotypesAPI")
        self.hk = hk_api
        self.gt = gt_api

    def data(self, analysis_obj: Analysis) -> dict:
        """Fetch data about an analysis to load genotypes.

        Returns: dict on form

        {
            "bcf": path_to_bcf,
            "samples_sex": [
                "sample_id: {
                    "pedigree": "male",
                    "analysis": "male"
                    }
            ]
        }

        """
        case_id = analysis_obj.case.internal_id
        LOG.info(f"Fetching upload genotype data for {case_id}")
        hk_version = self.hk.last_version(case_id)
        hk_bcf = self.get_bcf_file(hk_version)
        data = {"bcf": hk_bcf.full_path}
        if analysis_obj.pipeline in [Pipeline.BALSAMIC, Pipeline.BALSAMIC_UMI]:
            data["samples_sex"] = self._get_samples_sex_balsamic(case_obj=analysis_obj.case)
        elif analysis_obj.pipeline == Pipeline.MIP_DNA:
            data["samples_sex"] = self._get_samples_sex_mip(
                case_obj=analysis_obj.case, hk_version=hk_version
            )
        else:
            raise ValueError(f"Pipeline {analysis_obj.pipeline} does not support Genotype upload")
        return data

    def _get_samples_sex_mip(self, case_obj: Case, hk_version: housekeeper_models.Version) -> dict:
        qc_metrics_file = self.get_qcmetrics_file(hk_version)
        analysis_sexes = self.analysis_sex(qc_metrics_file)
        samples_sex = {}
        for link_obj in case_obj.links:
            sample_id = link_obj.sample.internal_id
            samples_sex[sample_id] = {
                "pedigree": link_obj.sample.sex,
                "analysis": analysis_sexes[sample_id],
            }
        return samples_sex

    def _get_samples_sex_balsamic(self, case_obj: Case) -> dict:
        samples_sex = {}
        for link_obj in case_obj.links:
            if link_obj.sample.is_tumour:
                continue
            sample_id = link_obj.sample.internal_id
            samples_sex[sample_id] = {
                "pedigree": link_obj.sample.sex,
                "analysis": Sex.UNKNOWN,
            }
        return samples_sex

    def analysis_sex(self, qc_metrics_file: Path) -> dict:
        """Fetch analysis sex for each sample of an analysis."""
        qc_metrics: MIPMetricsDeliverables = self.get_parsed_qc_metrics_data(qc_metrics_file)
        return {
            sample_id_metric.sample_id: sample_id_metric.predicted_sex
            for sample_id_metric in qc_metrics.sample_id_metrics
        }

    def get_bcf_file(self, hk_version_obj: housekeeper_models.Version) -> housekeeper_models.File:
        """Fetch a bcf file and return the file object"""
        genotype_files: list = self._get_genotype_files(version_id=hk_version_obj.id)
        for genotype_file in genotype_files:
            if self._is_variant_file(genotype_file=genotype_file):
                LOG.debug(f"Found bcf file {genotype_file.full_path}")
                return genotype_file
        raise FileNotFoundError(f"No vcf or bcf file found for bundle {hk_version_obj.bundle_id}")

    def get_qcmetrics_file(self, hk_version_obj: housekeeper_models.Version) -> Path:
        """Fetch a qc_metrics file and return the path"""
        hk_qcmetrics = self.hk.files(
            version=hk_version_obj.id, tags=HkMipAnalysisTag.QC_METRICS
        ).first()
        LOG.debug(f"Found qc metrics file {hk_qcmetrics.full_path}")
        return Path(hk_qcmetrics.full_path)

    @staticmethod
    def get_parsed_qc_metrics_data(qc_metrics: Path) -> MIPMetricsDeliverables:
        """Parse the information from a qc metrics file"""
        qcmetrics_raw: dict = ReadFile.get_content_from_file(
            file_format=FileFormat.YAML, file_path=qc_metrics
        )
        return MIPMetricsDeliverables(**qcmetrics_raw)

    def upload(self, data: dict, replace: bool = False):
        """Upload data about genotypes for a family of samples."""
        self.gt.upload(str(data["bcf"]), data["samples_sex"], force=replace)

    @staticmethod
    def _is_variant_file(genotype_file: housekeeper_models.File):
        return genotype_file.full_path.endswith("vcf.gz") or genotype_file.full_path.endswith("bcf")

    def _get_genotype_files(self, version_id: int) -> list:
        return self.hk.files(version=version_id, tags=["genotype"]).all()

    @staticmethod
    def is_suitable_for_genotype_upload(case_obj: Case) -> bool:
        """Check if a cancer case is contains WGS and normal sample."""

        samples: list[Sample] = case_obj.samples
        return any(
            (not sample.is_tumour and PrepCategory.WHOLE_GENOME_SEQUENCING == sample.prep_category)
            for sample in samples
        )



import datetime as dt
from typing import Iterable

from sqlalchemy.orm import Query

from cg.store import Store
from cg.store.filters.status_application_version_filters import (
    filter_application_versions_before_valid_from,
    filter_application_versions_by_application_entry_id,
    filter_application_versions_by_application_version_entry_id,
    order_application_versions_by_valid_from_desc,
)
from cg.store.models import Application, ApplicationVersion


def test_filter_application_version_by_application_entry_id_correct_id(
    base_store: Store,
):
    """Test that the correct application version is returned when using a correct application entry id."""
    # GIVEN a store with application versions
    app_version_query: Query = base_store._get_query(table=ApplicationVersion)

    # GIVEN an application in store different in id from at least another application in store
    application: Application = base_store.get_applications()[0]
    assert application.id != base_store.get_applications()[1].id

    # WHEN the application version query is filtered by the id of the application
    filtered_app_version_query: Query = filter_application_versions_by_application_entry_id(
        application_versions=app_version_query,
        application_entry_id=application.id,
    )

    # THEN the filtered query is shorter than the unfiltered query
    assert app_version_query.count() > filtered_app_version_query.count()

    # THEN the application id of all the filtered query entries is equal to the id used to filter
    application_ids: list[int] = [
        application_version.application_id
        for application_version in filtered_app_version_query.all()
    ]
    assert all(application_id == application.id for application_id in application_ids)


def test_filter_application_version_by_application_entry_id_wrong_id(
    base_store: Store,
    invalid_application_id: int,
):
    """Test that an empty query is returned when using an incorrect application id."""
    # GIVEN a store with application versions
    app_version_query: Query = base_store._get_query(table=ApplicationVersion)

    # WHEN filtering with an invalid application id
    filtered_app_version_query: Query = filter_application_versions_by_application_entry_id(
        application_versions=app_version_query,
        application_entry_id=invalid_application_id,
    )

    # THEN the filtered query is empty
    assert filtered_app_version_query.count() == 0


def test_filter_application_version_by_application_entry_id_empty_query(
    store: Store, application_id: int = 1
):
    """Test that filtering an empty query by application entry id returns an empty query."""
    # GIVEN a store without any application version
    app_version_query: Query = store._get_query(table=ApplicationVersion)
    assert app_version_query.count() == 0

    # WHEN filtering by application id
    filtered_app_version_query: Query = filter_application_versions_by_application_entry_id(
        application_versions=app_version_query,
        application_entry_id=application_id,
    )

    # THEN the function returns an empty query
    assert isinstance(filtered_app_version_query, Query)
    assert filtered_app_version_query.count() == 0


def test_filter_application_versions_before_valid_from_valid_date(
    base_store: Store,
):
    """Test that filtering by `valid_from` returns a query with older elements than the given date."""
    # GIVEN a store with application versions with different dates
    app_version_query: Query = base_store._get_query(table=ApplicationVersion).order_by(
        ApplicationVersion.valid_from
    )
    first_app_version: ApplicationVersion = app_version_query.first()
    third_app_version: ApplicationVersion = app_version_query.offset(2).first()
    assert first_app_version.valid_from < third_app_version.valid_from

    # WHEN filtering the application version query by `valid_from` of the third query
    filtered_app_version_query: Query = filter_application_versions_before_valid_from(
        application_versions=app_version_query,
        valid_from=third_app_version.valid_from,
    )

    # THEN a query with the two first application versions is returned
    assert filtered_app_version_query.count() == 2

    # THEN the date of the newest filtered application version is older than the filter date.
    newest_application_version: ApplicationVersion = filtered_app_version_query.order_by(
        ApplicationVersion.valid_from.desc()
    ).first()
    assert newest_application_version.valid_from < third_app_version.valid_from


def test_filter_application_versions_before_valid_from_future_date(
    base_store: Store,
    future_date: dt.datetime,
):
    """Test that filtering by `valid_from` with a future date returns the unfiltered query."""
    # GIVEN a store with application versions
    app_version_query: Query = base_store._get_query(table=ApplicationVersion)

    # WHEN filtering using a future date
    filtered_app_version_query: Query = filter_application_versions_before_valid_from(
        application_versions=app_version_query,
        valid_from=future_date,
    )

    # THEN the filtered query has the same elements as the unfiltered query
    app_version_pair: Iterable[tuple[ApplicationVersion, ApplicationVersion]] = zip(
        app_version_query.all(), filtered_app_version_query.all()
    )
    assert all(non_filtered == filtered for non_filtered, filtered in app_version_pair)


def test_filter_application_versions_before_valid_from_old_date(
    base_store: Store,
    old_timestamp: dt.datetime,
):
    """Test that filtering by `valid_from` with an old date returns an empty query."""
    # GIVEN a store with application versions
    app_version_query: Query = base_store._get_query(table=ApplicationVersion)

    # WHEN filtering using a past date
    filtered_app_version_query: Query = filter_application_versions_before_valid_from(
        application_versions=app_version_query,
        valid_from=old_timestamp,
    )

    # THEN the filtered query is empty
    assert filtered_app_version_query.count() == 0


def test_filter_application_version_before_valid_from_empty_query(
    store: Store,
    timestamp_now: dt.datetime,
):
    """Test that filtering an empty query by valid_from returns an empty query."""
    # GIVEN a store without any application version
    app_version_query: Query = store._get_query(table=ApplicationVersion)
    assert app_version_query.count() == 0

    # WHEN filtering by application id
    filtered_app_version_query: Query = filter_application_versions_before_valid_from(
        application_versions=app_version_query,
        valid_from=timestamp_now,
    )

    # THEN the function returns an empty query
    assert isinstance(filtered_app_version_query, Query)
    assert filtered_app_version_query.count() == 0


def test_order_application_versions_by_valid_from_desc(
    base_store: Store,
):
    """Test that ordering an application version query returns a query ordered by 'valid_from'."""
    # GIVEN a store with application versions with different dates
    app_version_query: Query = base_store._get_query(table=ApplicationVersion)

    # WHEN ordering the query by `valid_from`
    ordered_app_versions: list[ApplicationVersion] = list(
        order_application_versions_by_valid_from_desc(application_versions=app_version_query)
    )
    n_app_versions: int = len(ordered_app_versions)

    # THEN the elements of the ordered query have descending order of valid_from
    assert all(
        ordered_app_versions[i].valid_from > ordered_app_versions[i + 1].valid_from
        for i in range(n_app_versions - 1)
    )


def test_order_application_versions_by_valid_from_desc_empty_query(store: Store):
    """Test that ordering an empty query by valid_from returns an empty query."""
    # GIVEN a store without any application version
    app_version_query: Query = store._get_query(table=ApplicationVersion)
    assert app_version_query.count() == 0

    # WHEN filtering by application id
    ordered_app_version_query: Query = order_application_versions_by_valid_from_desc(
        application_versions=app_version_query
    )

    # THEN the function returns an empty query
    assert isinstance(ordered_app_version_query, Query)
    assert ordered_app_version_query.count() == 0


def test_filter_application_versions_by_application_version_entry_id_no_match(
    store_with_different_application_versions: Store,
):
    """Test that no application versions are returned when there is no matching application version entry id."""
    # GIVEN a store containing multiple application versions
    application_versions_query: Query = store_with_different_application_versions._get_query(
        table=ApplicationVersion
    )
    non_existent_application_version_entry_id = "non_existent_application_version_entry_id"

    # WHEN filtering application versions by a non-matching application version entry id
    filtered_application_versions: Query = (
        filter_application_versions_by_application_version_entry_id(
            application_versions=application_versions_query,
            application_version_entry_id=non_existent_application_version_entry_id,
        )
    )

    # THEN the query should return no application versions
    assert filtered_application_versions.count() == 0


def test_filter_application_versions_by_application_version_entry_id_match(
    store_with_different_application_versions: Store,
):
    """Test that application versions with matching application version entry id are returned."""
    # GIVEN a store containing multiple application versions
    application_versions_query: Query = store_with_different_application_versions._get_query(
        table=ApplicationVersion
    )
    existing_application_version_entry_id = application_versions_query.first().id

    # WHEN filtering application versions by a matching application version entry id
    filtered_application_versions: Query = (
        filter_application_versions_by_application_version_entry_id(
            application_versions=application_versions_query,
            application_version_entry_id=existing_application_version_entry_id,
        )
    )

    # THEN the query should return the application version with the matching entry id
    assert filtered_application_versions.count() == 1
    assert filtered_application_versions.first().id == existing_application_version_entry_id



import logging

from cg.apps.demultiplex.sample_sheet.sample_models import (
    FlowCellSampleBcl2Fastq,
    FlowCellSampleBCLConvert,
)
from cg.apps.demultiplex.sample_sheet.sample_sheet_creator import (
    SampleSheetCreatorBcl2Fastq,
    SampleSheetCreatorBCLConvert,
)
from cg.constants.demultiplexing import BclConverter
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData

LOG = logging.getLogger(__name__)


def get_sample_sheet_creator(
    flow_cell: FlowCellDirectoryData,
    lims_samples: list[FlowCellSampleBcl2Fastq | FlowCellSampleBCLConvert],
    force: bool,
) -> SampleSheetCreatorBcl2Fastq | SampleSheetCreatorBCLConvert:
    """Returns an initialised sample sheet creator according to the demultiplexing software."""
    if flow_cell.bcl_converter == BclConverter.BCL2FASTQ:
        return SampleSheetCreatorBcl2Fastq(
            flow_cell=flow_cell, lims_samples=lims_samples, force=force
        )
    return SampleSheetCreatorBCLConvert(flow_cell=flow_cell, lims_samples=lims_samples, force=force)


def create_sample_sheet(
    flow_cell: FlowCellDirectoryData,
    lims_samples: list[FlowCellSampleBcl2Fastq | FlowCellSampleBCLConvert],
    force: bool = False,
) -> list[list[str]]:
    """Create a sample sheet for a flow cell."""
    sample_sheet_creator: SampleSheetCreatorBcl2Fastq | SampleSheetCreatorBCLConvert = (
        get_sample_sheet_creator(
            flow_cell=flow_cell,
            lims_samples=lims_samples,
            force=force,
        )
    )
    LOG.info(
        f"Constructing a {flow_cell.bcl_converter} sample sheet "
        f"for the {flow_cell.sequencer_type} flow cell {flow_cell.id}"
    )
    return sample_sheet_creator.construct_sample_sheet()



"""Contains base models to be inherited from in other archive software files."""
from abc import abstractmethod

from housekeeper.store.models import File
from pydantic import BaseModel, ConfigDict

from cg.models.cg_config import DataFlowConfig
from cg.store.models import Sample


class FileAndSample(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    file: File
    sample: Sample


class FileTransferData(BaseModel):
    """Base class for classes representing files to be archived."""

    @classmethod
    @abstractmethod
    def create_from_file_and_sample(
        cls, file: File, sample: Sample, is_archiving: bool
    ) -> "FileTransferData":
        """Instantiates the class from a File and Sample object."""
        pass


class ArchiveHandler:
    """Base class for classes handling different archiving programs."""

    @abstractmethod
    def __init__(self, config: DataFlowConfig):
        """Initiates the ArchiveHandler based on the provided configuration."""
        pass

    @abstractmethod
    def archive_file(self, file_and_sample: FileAndSample):
        """Archives all folders provided, to their corresponding destination,
        as given by sources and destination parameter."""
        pass

    @abstractmethod
    def retrieve_files(self, files_and_samples: list[FileAndSample]):
        """Retrieves all files for all samples for the given flowcell."""
        pass

    @abstractmethod
    def convert_into_transfer_data(
        self, files_and_samples: list[FileAndSample], is_archiving: bool = True
    ) -> list[FileTransferData]:
        """Converts the provided files_and_samples into a list of objects formatted for the specific archiving flow."""
        pass

    @abstractmethod
    def is_job_done(self, job_id: int) -> bool:
        """Returns true if job has been completed, false otherwise."""
        pass

    @abstractmethod
    def delete_file(self, file_and_sample: FileAndSample) -> None:
        """Deletes a file at the archive location."""



"""Functions that handle files in the context of scout uploading"""
import logging

import requests
from housekeeper.store.models import File, Version

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.constants.housekeeper_tags import HK_DELIVERY_REPORT_TAG
from cg.meta.upload.scout.hk_tags import CaseTags, SampleTags
from cg.models.scout.scout_load_config import ScoutIndividual, ScoutLoadConfig
from cg.store.models import Analysis, CaseSample, Sample

LOG = logging.getLogger(__name__)


# Maps keys that are used in scout load config on tags that are used in scout


class ScoutConfigBuilder:
    """Base class for handling files that should be included in Scout upload"""

    def __init__(self, hk_version_obj: Version, analysis_obj: Analysis, lims_api: LimsAPI):
        self.hk_version_obj: Version = hk_version_obj
        self.analysis_obj: Analysis = analysis_obj
        self.lims_api: LimsAPI = lims_api
        self.case_tags: CaseTags
        self.sample_tags: SampleTags
        self.load_config: ScoutLoadConfig = ScoutLoadConfig(
            delivery_report=self.get_file_from_hk({HK_DELIVERY_REPORT_TAG})
        )

    def add_common_info_to_load_config(self) -> None:
        """Add the mandatory common information to a scout load config object"""
        self.load_config.analysis_date = self.analysis_obj.completed_at
        self.load_config.default_gene_panels = self.analysis_obj.case.panels
        self.load_config.family = self.analysis_obj.case.internal_id
        self.load_config.family_name = self.analysis_obj.case.name
        self.load_config.owner = self.analysis_obj.case.customer.internal_id
        self.load_config.synopsis = self.analysis_obj.case.synopsis
        self.include_cohorts()
        self.include_phenotype_groups()
        self.include_phenotype_terms()

    def add_common_sample_info(
        self, config_sample: ScoutIndividual, case_sample: CaseSample
    ) -> None:
        """Add the information to a sample that is common for different analysis types"""
        sample_id: str = case_sample.sample.internal_id
        LOG.info("Building sample %s", sample_id)
        lims_sample = dict()
        try:
            lims_sample = self.lims_api.sample(sample_id) or {}
        except requests.exceptions.HTTPError as ex:
            LOG.info("Could not fetch sample %s from LIMS: %s", sample_id, ex)

        config_sample.sample_id = sample_id
        config_sample.sex = case_sample.sample.sex
        config_sample.phenotype = case_sample.status
        config_sample.analysis_type = (
            case_sample.sample.application_version.application.analysis_type
        )
        config_sample.sample_name = case_sample.sample.name
        config_sample.tissue_type = lims_sample.get("source", "unknown")
        config_sample.subject_id = case_sample.sample.subject_id

    def add_common_sample_files(
        self,
        config_sample: ScoutIndividual,
        case_sample: CaseSample,
    ) -> None:
        """Add common sample files for different analysis types."""
        LOG.info(f"Adding common files for sample {case_sample.sample.internal_id}")
        self.include_sample_alignment_file(config_sample)
        self.include_sample_files(config_sample)

    def build_config_sample(self, case_sample: CaseSample) -> ScoutIndividual:
        """Build a sample for the scout load config"""
        raise NotImplementedError

    def build_load_config(self) -> ScoutLoadConfig:
        """Build a load config for uploading a case to scout"""
        raise NotImplementedError

    def include_sample_files(self, _config_sample: ScoutIndividual) -> None:
        """Include all files that are used on sample level in Scout"""
        return None

    def include_case_files(self) -> None:
        """Include all files that are used on case level in scout"""
        raise NotImplementedError

    def include_phenotype_terms(self) -> None:
        LOG.info("Adding phenotype terms to scout load config")
        phenotype_terms: set[str] = set()
        link_obj: CaseSample
        for link_obj in self.analysis_obj.case.links:
            sample_obj: Sample = link_obj.sample
            for phenotype_term in sample_obj.phenotype_terms:
                LOG.debug(
                    "Adding term %s from sample %s to phenotype terms",
                    phenotype_term,
                    sample_obj.internal_id,
                )
                phenotype_terms.add(phenotype_term)
        if phenotype_terms:
            self.load_config.phenotype_terms = list(phenotype_terms)

    def include_phenotype_groups(self) -> None:
        LOG.info("Adding phenotype groups to scout load config")
        phenotype_groups: set[str] = set()
        link_obj: CaseSample
        for link_obj in self.analysis_obj.case.links:
            sample_obj: Sample = link_obj.sample
            for phenotype_group in sample_obj.phenotype_groups:
                LOG.debug(
                    "Adding group %s from sample %s to phenotype groups",
                    phenotype_group,
                    sample_obj.internal_id,
                )
                phenotype_groups.add(phenotype_group)
        if phenotype_groups:
            self.load_config.phenotype_groups = list(phenotype_groups)

    def include_cohorts(self) -> None:
        LOG.info("Including cohorts to scout load config")
        cohorts: list[str] = self.analysis_obj.case.cohorts
        if cohorts:
            LOG.debug("Adding cohorts %s", ", ".join(cohorts))
            self.load_config.cohorts = cohorts

    def include_cnv_report(self) -> None:
        LOG.info("Include CNV report to case")
        self.load_config.cnv_report = self.get_file_from_hk(
            hk_tags=self.case_tags.cnv_report, latest=True
        )

    def include_multiqc_report(self) -> None:
        LOG.info("Include MultiQC report to case")
        self.load_config.multiqc = self.get_file_from_hk(
            hk_tags=self.case_tags.multiqc_report, latest=True
        )

    def include_sample_alignment_file(self, config_sample: ScoutIndividual) -> None:
        """Include the alignment file for a sample

        First add the bam.
        Cram is preferred so overwrite if found
        """
        sample_id: str = config_sample.sample_id
        config_sample.alignment_path = self.get_sample_file(
            hk_tags=self.sample_tags.bam_file, sample_id=sample_id
        )

        config_sample.alignment_path = self.get_sample_file(
            hk_tags=self.sample_tags.alignment_file, sample_id=sample_id
        )

    def get_sample_file(self, hk_tags: set[str], sample_id: str) -> str | None:
        """Return a file that is specific for a individual from housekeeper"""
        tags: set = hk_tags.copy()
        tags.add(sample_id)
        return self.get_file_from_hk(hk_tags=tags)

    def get_file_from_hk(self, hk_tags: set[str], latest: bool | None = False) -> str | None:
        """Get a file from housekeeper and return the path as a string."""
        LOG.info(f"Get file with tags {hk_tags}")
        if not hk_tags:
            LOG.debug("No tags provided, skipping")
            return None
        hk_file: File | None = (
            HousekeeperAPI.get_latest_file_from_version(version=self.hk_version_obj, tags=hk_tags)
            if latest
            else HousekeeperAPI.get_file_from_version(version=self.hk_version_obj, tags=hk_tags)
        )
        return hk_file.full_path if hk_file else None



"""Module for reading and writing xml files."""
import logging
from pathlib import Path
from xml.etree.ElementTree import Element, ElementTree, parse

from cg.constants import FileExtensions
from cg.exc import XMLError
from cg.io.validate_path import validate_file_suffix

LOG = logging.getLogger(__name__)


def read_xml(file_path: Path) -> ElementTree:
    """Read content in a xml file to an ElementTree."""
    validate_file_suffix(path_to_validate=file_path, target_suffix=FileExtensions.XML)
    tree = parse(file_path)
    return tree


def write_xml(tree: ElementTree, file_path: Path) -> None:
    """Write content to a xml file."""
    tree.write(file_path, encoding="utf-8", xml_declaration=True)


def validate_node_exists(node: Element | None, name: str) -> None:
    """Validates if the given node is not None.
    Raises:
        XMLError: If the node is None
    """
    if node is None:
        message = f"Could not find node with name {name} in XML tree"
        LOG.warning(message)
        raise XMLError(message)


def get_tree_node(tree: ElementTree, node_name: str) -> Element:
    """Return the node of a tree given its name if it exists."""
    xml_node: Element = tree.find(node_name)
    validate_node_exists(node=xml_node, name=node_name)
    return xml_node



import http
import json
import logging
import tempfile
from functools import wraps
from pathlib import Path
from typing import Any

import cachecontrol
import requests
from flask import Blueprint, abort, current_app, g, jsonify, make_response, request
from google.auth.transport import requests as google_requests
from google.oauth2 import id_token
from pydantic.v1 import ValidationError
from requests.exceptions import HTTPError
from sqlalchemy.exc import IntegrityError
from urllib3.exceptions import MaxRetryError, NewConnectionError
from werkzeug.utils import secure_filename

from cg.apps.orderform.excel_orderform_parser import ExcelOrderformParser
from cg.apps.orderform.json_orderform_parser import JsonOrderformParser
from cg.constants import ANALYSIS_SOURCES, METAGENOME_SOURCES
from cg.constants.constants import FileFormat
from cg.exc import OrderError, OrderFormError, TicketCreationError
from cg.io.controller import WriteStream
from cg.meta.orders import OrdersAPI
from cg.models.orders.order import OrderIn, OrderType
from cg.models.orders.orderform_schema import Orderform
from cg.server.ext import db, lims, osticket
from cg.store.models import (
    Analysis,
    Application,
    ApplicationLimitations,
    Case,
    Customer,
    Flowcell,
    Pool,
    Sample,
    SampleLaneSequencingMetrics,
    User,
)

LOG = logging.getLogger(__name__)
BLUEPRINT = Blueprint("api", __name__, url_prefix="/api/v1")


session = requests.session()
cached_session = cachecontrol.CacheControl(session)


def verify_google_token(token):
    request = google_requests.Request(session=cached_session)
    return id_token.verify_oauth2_token(id_token=token, request=request)


def is_public(route_function):
    @wraps(route_function)
    def public_endpoint(*args, **kwargs):
        return route_function(*args, **kwargs)

    public_endpoint.is_public = True
    return public_endpoint


@BLUEPRINT.before_request
def before_request():
    """Authorize API routes with JSON Web Tokens."""
    if not request.is_secure:
        return abort(
            make_response(
                jsonify(message="Only https requests accepted"), http.HTTPStatus.FORBIDDEN
            )
        )

    if request.method == "OPTIONS":
        return make_response(jsonify(ok=True), http.HTTPStatus.NO_CONTENT)

    endpoint_func = current_app.view_functions[request.endpoint]
    if getattr(endpoint_func, "is_public", None):
        return

    auth_header = request.headers.get("Authorization")
    if not auth_header:
        return abort(
            make_response(
                jsonify(message="no JWT token found on request"), http.HTTPStatus.UNAUTHORIZED
            )
        )

    jwt_token = auth_header.split("Bearer ")[-1]
    try:
        user_data = verify_google_token(jwt_token)
    except ValueError as e:
        LOG.error(f"Error occurred while decoding JWT token: {e}")
        return abort(
            make_response(
                jsonify(message="outdated login certificate"), http.HTTPStatus.UNAUTHORIZED
            )
        )

    user: User = db.get_user_by_email(user_data["email"])
    if user is None or not user.order_portal_login:
        message = f"{user_data['email']} doesn't have access"
        LOG.error(message)
        return abort(make_response(jsonify(message=message), http.HTTPStatus.FORBIDDEN))

    g.current_user = user


@BLUEPRINT.route("/submit_order/<order_type>", methods=["POST"])
def submit_order(order_type):
    """Submit an order for samples."""
    api = OrdersAPI(lims=lims, status=db, osticket=osticket)
    error_message: str
    try:
        request_json = request.get_json()
        LOG.info(
            "processing order: %s",
            WriteStream.write_stream_from_content(
                content=request_json, file_format=FileFormat.JSON
            ),
        )
        project: OrderType = OrderType(order_type)
        order_in: OrderIn = OrderIn.parse_obj(request_json, project=project)

        result = api.submit(
            project=project,
            order_in=order_in,
            user_name=g.current_user.name,
            user_mail=g.current_user.email,
        )
    except (  # user misbehaviour
        OrderError,
        OrderFormError,
        ValidationError,
        ValueError,
    ) as error:
        error_message = error.message if hasattr(error, "message") else str(error)
        http_error_response = http.HTTPStatus.BAD_REQUEST
        LOG.error(error_message)
    except (  # system misbehaviour
        AttributeError,
        ConnectionError,
        HTTPError,
        IntegrityError,
        KeyError,
        NewConnectionError,
        MaxRetryError,
        TimeoutError,
        TicketCreationError,
        TypeError,
    ) as error:
        LOG.exception(error)
        error_message = error.message if hasattr(error, "message") else str(error)
        http_error_response = http.HTTPStatus.INTERNAL_SERVER_ERROR
    else:
        return jsonify(
            project=result["project"], records=[record.to_dict() for record in result["records"]]
        )

    if error_message:
        return abort(make_response(jsonify(message=error_message), http_error_response))


@BLUEPRINT.route("/cases")
def get_cases():
    """Return cases with links for a customer from the database."""
    enquiry: str = request.args.get("enquiry")
    action: str = request.args.get("action")

    customers: list[Customer] = _get_current_customers()
    cases: list[Case] = _get_cases(enquiry=enquiry, action=action, customers=customers)

    nr_cases: int = len(cases)
    cases_with_links: list[dict] = [case.to_dict(links=True) for case in cases]
    return jsonify(families=cases_with_links, total=nr_cases)


def _get_current_customers() -> list[Customer] | None:
    """Return customers if the current user is not an admin."""
    return g.current_user.customers if not g.current_user.is_admin else None


def _get_cases(
    enquiry: str | None, action: str | None, customers: list[Customer] | None
) -> list[Case]:
    """Get cases based on the provided filters."""
    return db.get_cases_by_customers_action_and_case_search(
        case_search=enquiry,
        customers=customers,
        action=action,
    )


@BLUEPRINT.route("/cases/<case_id>")
def parse_case(case_id):
    """Return a case with links."""
    case: Case = db.get_case_by_internal_id(internal_id=case_id)
    if case is None:
        return abort(http.HTTPStatus.NOT_FOUND)
    if not g.current_user.is_admin and (case.customer not in g.current_user.customers):
        return abort(http.HTTPStatus.FORBIDDEN)
    return jsonify(**case.to_dict(links=True, analyses=True))


@BLUEPRINT.route("/families_in_collaboration")
def parse_families_in_collaboration():
    """Return cases in collaboration."""

    customer_internal_id = request.args.get("customer")
    pipeline = request.args.get("data_analysis")
    case_search_pattern = request.args.get("enquiry")

    customer = db.get_customer_by_internal_id(customer_internal_id=customer_internal_id)

    cases = db.get_cases_by_customer_pipeline_and_case_search(
        case_search=case_search_pattern,
        customer=customer,
        pipeline=pipeline,
    )

    case_dicts = [case.to_dict(links=True) for case in cases]
    return jsonify(families=case_dicts, total=len(cases))


@BLUEPRINT.route("/families_in_collaboration/<family_id>")
def parse_family_in_collaboration(family_id):
    """Return a family with links."""
    case: Case = db.get_case_by_internal_id(internal_id=family_id)
    customer: Customer = db.get_customer_by_internal_id(
        customer_internal_id=request.args.get("customer")
    )
    if case.customer not in customer.collaborators:
        return abort(http.HTTPStatus.FORBIDDEN)
    return jsonify(**case.to_dict(links=True, analyses=True))


@BLUEPRINT.route("/samples")
def parse_samples():
    """Return samples."""
    if request.args.get("status") and not g.current_user.is_admin:
        return abort(http.HTTPStatus.FORBIDDEN)
    if request.args.get("status") == "incoming":
        samples: list[Sample] = db.get_samples_to_receive()
    elif request.args.get("status") == "labprep":
        samples: list[Sample] = db.get_samples_to_prepare()
    elif request.args.get("status") == "sequencing":
        samples: list[Sample] = db.get_samples_to_sequence()
    else:
        customers: list[Customer] | None = (
            None if g.current_user.is_admin else g.current_user.customers
        )
        samples: list[Sample] = db.get_samples_by_customer_id_and_pattern(
            pattern=request.args.get("enquiry"), customers=customers
        )
    limit = int(request.args.get("limit", 50))
    parsed_samples: list[dict] = [sample.to_dict() for sample in samples[:limit]]
    return jsonify(samples=parsed_samples, total=len(samples))


@BLUEPRINT.route("/samples_in_collaboration")
def parse_samples_in_collaboration():
    """Return samples in a customer group."""
    customer: Customer = db.get_customer_by_internal_id(
        customer_internal_id=request.args.get("customer")
    )
    samples: list[Sample] = db.get_samples_by_customer_id_and_pattern(
        pattern=request.args.get("enquiry"), customers=customer.collaborators
    )
    limit = int(request.args.get("limit", 50))
    parsed_samples: list[dict] = [sample.to_dict() for sample in samples[:limit]]
    return jsonify(samples=parsed_samples, total=len(samples))


@BLUEPRINT.route("/samples/<sample_id>")
def parse_sample(sample_id):
    """Return a single sample."""
    sample: Sample = db.get_sample_by_internal_id(sample_id)
    if sample is None:
        return abort(http.HTTPStatus.NOT_FOUND)
    if not g.current_user.is_admin and (sample.customer not in g.current_user.customers):
        return abort(http.HTTPStatus.FORBIDDEN)
    return jsonify(**sample.to_dict(links=True, flowcells=True))


@BLUEPRINT.route("/samples_in_collaboration/<sample_id>")
def parse_sample_in_collaboration(sample_id):
    """Return a single sample."""
    sample: Sample = db.get_sample_by_internal_id(sample_id)
    customer: Customer = db.get_customer_by_internal_id(
        customer_internal_id=request.args.get("customer")
    )
    if sample.customer not in customer.collaborators:
        return abort(http.HTTPStatus.FORBIDDEN)
    return jsonify(**sample.to_dict(links=True, flowcells=True))


@BLUEPRINT.route("/pools")
def parse_pools():
    """Return pools."""
    customers: list[Customer] | None = (
        g.current_user.customers if not g.current_user.is_admin else None
    )
    pools: list[Pool] = db.get_pools_to_render(
        customers=customers, enquiry=request.args.get("enquiry")
    )
    parsed_pools: list[dict] = [pool_obj.to_dict() for pool_obj in pools[:30]]
    return jsonify(pools=parsed_pools, total=len(pools))


@BLUEPRINT.route("/pools/<pool_id>")
def parse_pool(pool_id):
    """Return a single pool."""
    pool: Pool = db.get_pool_by_entry_id(entry_id=pool_id)
    if pool is None:
        return abort(http.HTTPStatus.NOT_FOUND)
    if not g.current_user.is_admin and (pool.customer not in g.current_user.customers):
        return abort(http.HTTPStatus.FORBIDDEN)
    return jsonify(**pool.to_dict())


@BLUEPRINT.route("/flowcells")
def parse_flow_cells() -> Any:
    """Return flow cells."""
    flow_cells: list[Flowcell] = db.get_flow_cell_by_name_pattern_and_status(
        flow_cell_statuses=[request.args.get("status")],
        name_pattern=request.args.get("enquiry"),
    )
    parsed_flow_cells: list[dict] = [flow_cell.to_dict() for flow_cell in flow_cells[:50]]
    return jsonify(flowcells=parsed_flow_cells, total=len(flow_cells))


@BLUEPRINT.route("/flowcells/<flowcell_id>")
def parse_flow_cell(flowcell_id):
    """Return a single flowcell."""
    flow_cell: Flowcell = db.get_flow_cell_by_name(flow_cell_name=flowcell_id)
    if flow_cell is None:
        return abort(http.HTTPStatus.NOT_FOUND)
    return jsonify(**flow_cell.to_dict(samples=True))


@BLUEPRINT.route("/flowcells/<flow_cell_name>/sequencing_metrics", methods=["GET"])
def get_sequencing_metrics(flow_cell_name: str):
    """Return sample lane sequencing metrics for a flow cell."""

    if not flow_cell_name:
        return jsonify({"error": "Invalid or missing flow cell id"}), http.HTTPStatus.BAD_REQUEST

    sequencing_metrics: list[
        SampleLaneSequencingMetrics
    ] = db.get_sample_lane_sequencing_metrics_by_flow_cell_name(flow_cell_name)

    if not sequencing_metrics:
        return (
            jsonify({"error": f"Sequencing metrics not found for flow cell {flow_cell_name}."}),
            http.HTTPStatus.NOT_FOUND,
        )

    return jsonify([metric.to_dict() for metric in sequencing_metrics])


@BLUEPRINT.route("/analyses")
def parse_analyses():
    """Return analyses."""
    if request.args.get("status") == "delivery":
        analyses: list[Analysis] = db.get_analyses_to_deliver_for_pipeline()
    elif request.args.get("status") == "upload":
        analyses: list[Analysis] = db.get_analyses_to_upload()
    else:
        analyses: list[Analysis] = db.get_analyses()
    parsed_analysis: list[dict] = [analysis_obj.to_dict() for analysis_obj in analyses[:30]]
    return jsonify(analyses=parsed_analysis, total=len(analyses))


@BLUEPRINT.route("/options")
def parse_options():
    """Return various options."""
    customers: list[Customer | None] = (
        db.get_customers() if g.current_user.is_admin else g.current_user.customers
    )

    app_tag_groups: dict[str, list[str]] = {"ext": []}
    applications: list[Application] = db.get_applications_is_not_archived()
    for application in applications:
        if not application.versions:
            LOG.debug(f"Skipping application {application} that doesn't have a price")
            continue
        if application.is_external:
            app_tag_groups["ext"].append(application.tag)
        if application.prep_category not in app_tag_groups:
            app_tag_groups[application.prep_category]: list[str] = []
        app_tag_groups[application.prep_category].append(application.tag)

    source_groups = {"metagenome": METAGENOME_SOURCES, "analysis": ANALYSIS_SOURCES}

    return jsonify(
        applications=app_tag_groups,
        beds=[bed.name for bed in db.get_active_beds()],
        customers=[
            {
                "text": f"{customer.name} ({customer.internal_id})",
                "value": customer.internal_id,
                "isTrusted": customer.is_trusted,
            }
            for customer in customers
        ],
        organisms=[
            {
                "name": organism.name,
                "reference_genome": organism.reference_genome,
                "internal_id": organism.internal_id,
                "verified": organism.verified,
            }
            for organism in db.get_all_organisms()
        ],
        panels=[panel.abbrev for panel in db.get_panels()],
        sources=source_groups,
    )


@BLUEPRINT.route("/me")
def parse_current_user_information():
    """Return information about current user."""
    if not g.current_user.is_admin and not g.current_user.customers:
        LOG.error(
            "%s is not admin and is not connected to any customers, aborting", g.current_user.email
        )
        return abort(http.HTTPStatus.FORBIDDEN)

    return jsonify(user=g.current_user.to_dict())


@BLUEPRINT.route("/applications")
@is_public
def parse_applications():
    """Return application tags."""
    applications: list[Application] = db.get_applications_is_not_archived()
    parsed_applications: list[dict] = [application.to_dict() for application in applications]
    return jsonify(applications=parsed_applications)


@BLUEPRINT.route("/applications/<tag>")
@is_public
def parse_application(tag: str):
    """Return an application tag."""
    application: Application = db.get_application_by_tag(tag=tag)
    if not application:
        return abort(
            make_response(jsonify(message="Application not found"), http.HTTPStatus.NOT_FOUND)
        )
    return jsonify(**application.to_dict())


@BLUEPRINT.route("/applications/<tag>/pipeline_limitations")
@is_public
def get_application_pipeline_limitations(tag: str):
    """Return application pipeline specific limitations."""
    application_limitations: list[ApplicationLimitations] = db.get_application_limitations_by_tag(
        tag
    )
    if not application_limitations:
        return jsonify(message="Application limitations not found"), http.HTTPStatus.NOT_FOUND
    return jsonify([limitation.to_dict() for limitation in application_limitations])


@BLUEPRINT.route("/orderform", methods=["POST"])
def parse_orderform():
    """Parse an orderform/JSON export."""
    input_file = request.files.get("file")
    filename = secure_filename(input_file.filename)

    error_message: str
    try:
        if filename.lower().endswith(".xlsx"):
            temp_dir = Path(tempfile.gettempdir())
            saved_path = str(temp_dir / filename)
            input_file.save(saved_path)
            order_parser = ExcelOrderformParser()
            order_parser.parse_orderform(excel_path=saved_path)
        else:
            json_data = json.load(input_file.stream, strict=False)
            order_parser = JsonOrderformParser()
            order_parser.parse_orderform(order_data=json_data)
        parsed_order: Orderform = order_parser.generate_orderform()
    except (  # user misbehaviour
        AttributeError,
        OrderFormError,
        OverflowError,
        ValidationError,
        ValueError,
    ) as error:
        error_message = error.message if hasattr(error, "message") else str(error)
        LOG.error(error_message)
        http_error_response = http.HTTPStatus.BAD_REQUEST
    except (  # system misbehaviour
        NewConnectionError,
        MaxRetryError,
        TimeoutError,
        TypeError,
    ) as error:
        LOG.exception(error)
        error_message = error.message if hasattr(error, "message") else str(error)
        http_error_response = http.HTTPStatus.INTERNAL_SERVER_ERROR
    else:
        return jsonify(**parsed_order.model_dump())

    if error_message:
        return abort(make_response(jsonify(message=error_message), http_error_response))



from click.testing import CliRunner

from cg.cli.get import get
from cg.constants import EXIT_SUCCESS
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import Customer
from tests.store_helpers import StoreHelpers


def test_get_case_by_name(cli_runner: CliRunner, base_context: CGConfig, helpers: StoreHelpers):
    # GIVEN A database with a customer in it
    status_db: Store = base_context.status_db
    customer: Customer = helpers.ensure_customer(store=status_db)

    # WHEN trying to get a non-existing case by name
    result = cli_runner.invoke(
        get,
        [
            "case",
            "-c",
            customer.internal_id,
            "-n",
            "dummy-case-name",
        ],
        obj=base_context,
    )

    # THEN it should not crash
    assert result.exit_code == EXIT_SUCCESS
